use std::hash::Hash;

use anyhow::{anyhow, Error};
use i2p::net::{I2pAddr, I2pSocketAddr};
use regex::Regex;
use serde::{Deserialize, Serialize};

/// Uniquely identifies a peer. The I2pAddr inside `addr` MUST be the peer's b32 address.
#[derive(Serialize, Deserialize, Debug, Clone, PartialEq, Eq, PartialOrd, Ord, Hash)]
pub struct PeerId {
    addr: I2pSocketAddr,
}

impl PeerId {
    pub fn try_from_b32(addr: &str, port: Option<u16>) -> anyhow::Result<Self> {
        let b32_regex = Regex::new(r"[abcdefghijklmnopqrstuvwxyz234567]{52}.b32.i2p").unwrap();
        if b32_regex.is_match(addr) {
            Ok(PeerId {
                addr: I2pSocketAddr::new(I2pAddr::new(addr), port.unwrap_or(0)),
            })
        } else {
            Err(Error::msg("The supplied address was not b32-formatted."))
        }
    }
    pub fn try_from_dest(dest: &str, port: Option<u16>) -> anyhow::Result<Self> {
        let b32 = I2pAddr::from_b64(dest).map_err(|e| anyhow!(e))?;
        Ok(PeerId {
            addr: I2pSocketAddr::new(b32, port.unwrap_or(0)),
        })
    }

    pub fn addr(&self) -> I2pSocketAddr {
        self.addr.to_owned()
    }

    pub fn addr_ref(&self) -> &I2pSocketAddr {
        &self.addr
    }
}

impl ToString for PeerId {
    fn to_string(&self) -> String {
        self.addr.to_string()
    }
}

impl From<PeerId> for I2pSocketAddr {
    fn from(value: PeerId) -> Self {
        value.addr
    }
}

impl From<I2pSocketAddr> for PeerId {
    fn from(value: I2pSocketAddr) -> Self {
        Self::try_from_b32(&value.dest().string(), Some(value.port())).unwrap_or(
            Self::try_from_dest(&value.dest().string(), Some(value.port())).expect(&format!(
                "I2pSocketAddr was neither in b32 nor in b64 format: {value:?}"
            )),
        )
    }
}

impl Default for PeerId {
    fn default() -> Self {
        PeerId {
            addr: I2pSocketAddr::new(I2pAddr::new(""), 0),
        }
    }
}

#[cfg(test)]
mod tests {
    use super::PeerId;

    #[test]
    fn from_b32() {
        let addr = "abcdefghijklmnopqrstuvwxyz234567abcdefghijklmnopqrst.b32.i2p";
        let peer_id = PeerId::try_from_b32(addr, None);

        assert!(peer_id.is_ok())
    }
}
